from importlib import resources as impresources

from pydrake.geometry import HalfSpace
from pydrake.math import RollPitchYaw, RigidTransform, RotationMatrix
from pydrake.multibody.math import SpatialVelocity
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant, CoulombFriction
from pydrake.systems.framework import Diagram, DiagramBuilder, LeafSystem, PortDataType, BasicVector

from manipulation.scenarios import AddMultibodyTriad

import numpy as np

# Internal Imports
from brom_drake import example_helpers as eh


class BlockHandlerSystem(LeafSystem):
    def __init__(
        self,
        plant: MultibodyPlant,
        scene_graph,
        block_name: str = "block_with_slots",
        model_urdf_path: str = None,
    ):
        LeafSystem.__init__(self)

        # Input Processing
        if model_urdf_path is None:
            model_urdf_path = str(
                impresources.files(eh) / "models/slider-block.urdf",
            )

        # Constants
        self.block_name = block_name

        # Add the Block to the given plant
        self.plant = plant
        self.block_model_idx = Parser(plant=self.plant).AddModels(model_urdf_path)[0]
        self.block_model_name = self.plant.GetModelInstanceName(self.block_model_idx)
        self.block_body_name = "block"

        AddGround(self.plant) #Add ground to plant

        # Add the Triad
        self.scene_graph = scene_graph
        AddMultibodyTriad(plant.GetFrameByName(self.block_body_name), self.scene_graph)

        self.plant.Finalize()
        self.context = self.plant.CreateDefaultContext()

        # Create Input Port for the Slider Block System
        self.desired_pose_port = self.DeclareVectorInputPort(
            "desired_pose",
            BasicVector(6),
        )

        # Create Output Port which should share the pose of the block
        self.DeclareVectorOutputPort(
                "measured_block_pose",
                BasicVector(6),
                self.SetBlockPose,
                {self.time_ticket()}   # indicate that this doesn't depend on any inputs,
                )                      # but should still be updated each timestep

    def SetBlockPose(self, context, output):
        """
        Description:
            This function sets the desired pose of the block.
        """

        # Get Desired Pose from Port
        plant_context = self.context
        pose_as_vec = self.desired_pose_port.Eval(context)

        self.plant.SetFreeBodyPose(
            plant_context,
            self.plant.GetBodyByName(self.block_body_name),
            RigidTransform(RollPitchYaw(pose_as_vec[:3]),pose_as_vec[3:])
        )

        self.plant.SetFreeBodySpatialVelocity(
            self.plant.GetBodyByName(self.block_body_name),
            SpatialVelocity(np.zeros(3),np.array([0.0,0.0,0.0])),
            plant_context
            )

        X_WBlock = self.plant.GetFreeBodyPose(
            plant_context,
            self.plant.GetBodyByName(self.block_body_name)
        )

        pose_as_vector = np.hstack([RollPitchYaw(X_WBlock.rotation()).vector(), X_WBlock.translation()])

        # Create Output
        output.SetFromVector(pose_as_vector)

    def SetInitialBlockState(self,diagram_context):
        """
        Description:
            Sets the initial position to be slightly above the ground (small, positive z value)
            to be .
        """

        # Set Pose
        p_WBlock = [0.0, 0.0, 0.2]
        R_WBlock = RotationMatrix.MakeXRotation(np.pi/2.0) # RotationMatrix.MakeXRotation(-np.pi/2.0)
        X_WBlock = RigidTransform(R_WBlock, p_WBlock)
        self.plant.SetFreeBodyPose(
            self.plant.GetMyContextFromRoot(diagram_context),
            self.plant.GetBodyByName(self.block_body_name),
            X_WBlock)

        # Set Velocities
        self.plant.SetFreeBodySpatialVelocity(
            self.plant.GetBodyByName(self.block_body_name),
            SpatialVelocity(np.zeros(3),np.array([0.0,0.0,0.0])),
            self.plant.GetMyContextFromRoot(diagram_context))

def AddGround(plant):
    """
    Add a flat ground with friction
    """

    # Constants
    transparent_color = np.array([0.5, 0.5, 0.5, 0])
    nontransparent_color = np.array([0.5, 0.5, 0.5, 0.1])

    p_GroundOrigin = [0, 0.0, 0.0]
    R_GroundOrigin = RotationMatrix.MakeXRotation(0.0)
    X_GroundOrigin = RigidTransform(R_GroundOrigin,p_GroundOrigin)

    # Set Up Ground on Plant

    surface_friction = CoulombFriction(
            static_friction = 0.7,
            dynamic_friction = 0.5)
    plant.RegisterCollisionGeometry(
            plant.world_body(),
            X_GroundOrigin,
            HalfSpace(),
            "ground_collision",
            surface_friction)
    plant.RegisterVisualGeometry(
            plant.world_body(),
            X_GroundOrigin,
            HalfSpace(),
            "ground_visual",
            transparent_color)  # transparent


"""
Description:
    This file defines a Role class that is used to populate the
    scene with ONLY systems that match the appropriate structure/signature.
"""
from dataclasses import dataclass

from pydrake.systems.framework import DiagramBuilder

# Internal Imports
from brom_drake.utils import (
    find_all_systems_with_output_port, find_all_systems_with_input_port,
    Performer,
)

@dataclass
class PortPairing:
    external_port_name: str
    performer_port_name: str

@dataclass
class Role:
    name: str
    description: str
    required_input_definitions: list[PortPairing]
    required_output_definitions: list[PortPairing]

    def __str__(self):
        return self.name

    def connect_performer_to_system(
        self,
        builder: DiagramBuilder,
        performer: Performer,
    ):
        """
        Description
        -----------
        This method connects the performer to the system.
        :param builder: A DiagramBuilder object.
        :param performer: A system or diagram that should have the appropriate ports.
        :return:
        """
        # Setup

        # Check that the performer has the required input and output ports
        for port_pairing in self.required_input_definitions:
            assert performer.HasInputPort(port_pairing.performer_port_name), \
                f"Performer does not have input port \"{port_pairing.performer_port_name}\""

        for port_pairing in self.required_output_definitions:
            assert performer.HasOutputPort(port_pairing.performer_port_name), \
                f"Performer does not have output port \"{port_pairing.performer_port_name}\""

        # Connect the performer's inputs to other systems in the builder
        for port_pairing in self.required_input_definitions:
            performer_port = performer.GetInputPort(port_pairing.performer_port_name)

            # Get the external system that has the output port with the correct name
            systems_list = find_all_systems_with_output_port(builder, port_pairing.external_port_name)
            assert len(systems_list) == 1, \
                f"Expected 1 system to have port \"{port_pairing.external_port_name}\"," + \
                f" but found {len(systems_list)} systems with that output port."

            builder.Connect(
                systems_list[0].GetOutputPort(port_pairing.external_port_name),
                performer_port,
            )

        # Connect the performer's outputs to other systems in the builder
        for port_pairing in self.required_output_definitions:
            performer_port = performer.GetOutputPort(port_pairing.external_port_name)

            # Get the external system that has the input port with the correct name
            systems_list = find_all_systems_with_input_port(builder, port_pairing.external_port_name)
            assert len(systems_list) == 1, \
                f"Expected 1 system to have port \"{port_pairing.external_port_name}\"," + \
                f" but found {len(systems_list)} systems with that output port."

            builder.Connect(
                performer_port,
                systems_list[0].GetInputPort(port_pairing.external_port_name),
            )



from .end_effector_wrench_calculator import EndEffectorWrenchCalculator
from .gripper_type import GripperType
from .ur10e_station import UR10eStation

__all__ = [
    "EndEffectorWrenchCalculator",
    "GripperType",
    "UR10eStation",
]



from pathlib import Path
from typing import Union, List, Tuple

from pydrake.systems.analysis import Simulator
from pydrake.systems.framework import DiagramBuilder, Diagram, Context

#Internal Imports
from brom_drake.all import add_watcher_and_build
from .show_me_system import ShowMeSystem
from .show_me_this_model import ShowMeThisModel

def show_me_this_model_in_sim(
    path_to_model: Union[str, Path],
    desired_joint_positions: List[float],
    time_step: float = 1e-3,
    base_link_name: str = None,
)-> Tuple[ShowMeThisModel, Diagram, Context, Simulator]:
    # Setup
    builder = DiagramBuilder()

    # Define Scene
    scene = ShowMeThisModel(
        path_to_model=str(path_to_model),
        desired_joint_positions=desired_joint_positions,
        base_link_name=base_link_name,
        time_step=time_step,
    )

    diagram = scene.cast_scene_and_build(builder)
    diagram_context = diagram.CreateDefaultContext()

    # Set up simulation
    simulator = Simulator(diagram)
    simulator_context = simulator.get_mutable_context()
    scene.show_me_system.mutable_plant_context = scene.plant.GetMyMutableContextFromRoot(
        simulator_context,
    )

    return scene, diagram, diagram_context, simulator



from enum import IntEnum


class ArmControlMode(IntEnum):
    """
    Description:
        An enumeration of the different control modes that the arm can be in.
    """
    kJoint = 0
    kEndEffector = 1


from brom_drake.DiagramTarget import DiagramTarget
from brom_drake.DiagramWatcher import DiagramWatcher
from .add_watcher import add_watcher, add_watcher_and_build, parse_list_of_simplified_targets
from brom_drake.PortWatcher import PortWatcher, PortFigureArrangement, PortWatcherOptions
from brom_drake.example_helpers import BlockHandlerSystem
from brom_drake.urdf import DrakeReadyURDFConverter, drakeify_my_urdf

__all__ = [
    'DiagramTarget', 'DiagramWatcher',
    'add_watcher', 'add_watcher_and_build', 'parse_list_of_simplified_targets',
    'PortWatcher', 'PortFigureArrangement', 'PortWatcherOptions',
    'BlockHandlerSystem',
    'DrakeReadyURDFConverter', 'drakeify_my_urdf',
]


from pydrake.geometry import SceneGraph
from pydrake.systems.primitives import (
    VectorLogSink, ConstantVectorSource,
)

INELIGIBLE_SYSTEM_TYPES = [SceneGraph, VectorLogSink, ConstantVectorSource]

DEFAULT_PLOT_DIR = "./brom/watcher_plots"


from .constants import DEFAULT_PLOT_DIR
from .DiagramWatcher import DiagramWatcher
from .errors import UnrecognizedTargetError

__all__ = [
    'DiagramWatcher', 'UnrecognizedTargetError',
    'DEFAULT_PLOT_DIR',
]



from typing import Union, Tuple, List

from pydrake.systems.framework import DiagramBuilder, LeafSystem, Diagram

# Internal Imports
from brom_drake.scenes.roles import Role
from brom_drake.scenes.ids import SceneID
from brom_drake.utils import Performer

class BaseScene:
    """
    Base class for all scenes.
    """
    def __init__(self, **kwargs):
        pass

    def add_all_secondary_cast_members_to_builder(self, builder: DiagramBuilder):
        """
        Description
        -----------
        This method must be implemented by the subclass. It should add all
        elements to the builder.
        :param builder:
        :return:
        """
        pass

    def suggested_roles(self) -> List[Role]:
        """
        Description
        -----------
        This method should be implemented by the subclass.
        It should return a list of roles that are suggested for the scene.
        :return:
        """
        return []

    def fill_role(
        self,
        role: Role,
        system: Performer,
        builder: DiagramBuilder,
    ):
        """
        Description
        -----------
        This method should be implemented by the subclass. It should add the
        system to the role.
        :param role:
        :param system:
        :param builder:
        :return:
        """
        # Call the member method of the role object
        role.connect_performer_to_system(builder, system)

    def cast_scene(
        self,
        builder: DiagramBuilder,
        cast: Tuple[Role, Performer] = [],
    ):
        # Setup

        # Add all elements to the builder
        self.add_all_secondary_cast_members_to_builder(builder)

        # Fulfill each role-performer pair in the casting_call list
        for role, performer in cast:
            self.fill_role(role, performer, builder)

    def cast_scene_and_build(
        self,
        builder: DiagramBuilder,
        cast: Tuple[Role, Performer] = [],
    ) -> Diagram:
        # Setup
        self.cast_scene(builder, cast)

        # Build the diagram
        return builder.Build()

    @property
    def id(self) -> SceneID:
        """
        Description
        -----------
        The unique identifier for the scene
        :return:
        """
        return SceneID.kNotDefined


"""
gripper_controller.py
Description:

        This file defines the GripperController class.
"""
# External Imports
import importlib.resources as impresources

from pydrake.common.value import AbstractValue
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import JacobianWrtVariable
from pydrake.systems.framework import LeafSystem
from pydrake.systems.framework import BasicVector

import os

import numpy as np

# Internal Imports
from .gripper_target import GripperTarget
from brom_drake.robots.gripper_type import GripperType

from .. import robots


# Constants
package_dir = os.path.dirname(os.path.abspath(__file__))


class GripperController(LeafSystem):
    """
    A simple gripper controller with two modes: position and velocity.
    Both modes are essentially simple PD controllers.

                            -------------------------
                            |                       |
                            |                       |
    gripper_target -------> |   GripperController   | ---> applied_gripper_torques
    gripper_target_type --> |                       |
                            |                       | ---> measured_gripper_position
                            |                       | ---> measured_gripper_velocity
    gripper_state --------> |                       |
                            |                       |
                            |                       |
                            -------------------------
    """
    def __init__(self, gripper_type: GripperType):
        """
        Description
        -----------
        Constructor for the GripperController class.
        :param gripper_type: The type can be "Robotiq_2f_85" or "NoGripper"
        """
        LeafSystem.__init__(self)
        self.type = gripper_type
        self.context = None

        # State input port size depends on what type of gripper we're using
        if self.type == GripperType.Robotiq_2f_85:
            state_size = 12

            # We'll create a simple model of the gripper which is welded to the floor.
            # This will allow us to compute the distance between fingers.
            self.plant = MultibodyPlant(time_step=1.0) # time step doesn't matter
            gripper_urdf = str(
                impresources.files(robots) / "models/2f_85_gripper/urdf/robotiq_2f_85.urdf"
            )
            self.gripper = Parser(plant=self.plant).AddModels(gripper_urdf)[0]
            self.plant.WeldFrames(
                self.plant.world_frame(),
                self.plant.GetFrameByName("robotiq_arg2f_base_link"),
            )
            # Finalize the plant used for the gripper
            self.plant.Finalize()
            self.context = self.plant.CreateDefaultContext()

        else:
            raise RuntimeError("Invalid gripper type: %s" % self.type)

        # Declare input ports
        self.target_port = self.DeclareVectorInputPort("gripper_target", BasicVector(1))
        self.target_type_port = self.DeclareAbstractInputPort(
            "gripper_target_type",
            AbstractValue.Make(GripperTarget.kPosition),
        )
        self.state_port = self.DeclareVectorInputPort("gripper_state", BasicVector(state_size))

        # Declare output ports
        self.DeclareVectorOutputPort("applied_gripper_torque", BasicVector(2), self.CalcGripperTorque)
        self.DeclareVectorOutputPort(
            "measured_gripper_position", BasicVector(1), self.CalcGripperPosition,
            {self.time_ticket()}    # indicate that this doesn't depend on any inputs,
        )                           # but should still be updated each timestep
        self.DeclareVectorOutputPort(
            "measured_gripper_velocity", BasicVector(1), self.CalcGripperVelocity,
            {self.time_ticket()}    # indicate that this doesn't depend on any inputs,
        )                           # but should still be updated each timestep

    def ComputePosition(self, state):
        """
        Compute the gripper position from state data.
        This is especially useful for the 2F-85 gripper, since the
        state does not map neatly to the finger positions.
        """
        # Setup
        finger_position = None

        # Get the finger positions
        if self.type == GripperType.Robotiq_2f_85:
            # For the more complex 2F-85 gripper, we need to do some kinematics
            # calculations to figure out the gripper position
            self.plant.SetPositionsAndVelocities(self.context, state)

            right_finger = self.plant.GetFrameByName("right_inner_finger_pad")
            left_finger = self.plant.GetFrameByName("left_inner_finger_pad")
            base = self.plant.GetFrameByName("robotiq_arg2f_base_link")

            X_lf = self.plant.CalcRelativeTransform(self.context,
                                                    left_finger,
                                                    base)
            X_rf = self.plant.CalcRelativeTransform(self.context,
                                                    right_finger,
                                                    base)

            lf_pos = -X_lf.translation()[1]
            rf_pos = -X_rf.translation()[1]

            finger_position = np.array([lf_pos, rf_pos])

        else:
            raise NotImplementedError("Gripper type %s does not have ComputePosition() implemented" % self.type)

        return finger_position

    def ComputeVelocity(self, state):
        """
        Compute the gripper velocity from state data.
        This is especially useful for the 2F-85 gripper, since the
        state does not map neatly to the finger positions.
        """
        if self.type == GripperType.Robotiq_2f_85:
            # For the more complex 2F-85 gripper, we need to do some kinematics
            self.plant.SetPositionsAndVelocities(self.context, state)
            v = state[-self.plant.num_velocities():]

            right_finger = self.plant.GetFrameByName("right_inner_finger_pad")
            left_finger = self.plant.GetFrameByName("left_inner_finger_pad")
            base = self.plant.GetFrameByName("robotiq_arg2f_base_link")

            J_lf = self.plant.CalcJacobianTranslationalVelocity(
                self.context, JacobianWrtVariable.kV, left_finger, np.zeros(3), base, base,
            )
            J_rf = self.plant.CalcJacobianTranslationalVelocity(
                self.context, JacobianWrtVariable.kV, right_finger, np.zeros(3), base, base,
            )

            lf_vel = -(J_lf @ v)[1]
            rf_vel = (J_rf @ v)[1]

            finger_velocity = np.array([lf_vel, rf_vel])

        else:
            raise NotImplementedError("Gripper type %s does not have ComputeVelocity() implemented" % self.type)

        return finger_velocity

    def CalcGripperPosition(self, context, output):
        state = self.state_port.Eval(context)

        if self.type == "hande":
            width = 0.03
        elif self.type == GripperType.Robotiq_2f_85:  #2f_85
            width = 0.06
        else:
            raise NotImplementedError("Gripper type %s does not have CalcGripperPosition() implemented" % self.type)

        # Send a single number to match the hardware
        both_finger_positions = self.ComputePosition(state)
        net_position = 1/width* np.mean(both_finger_positions)

        output.SetFromVector([net_position])

    def CalcGripperVelocity(self, context, output):
        state = self.state_port.Eval(context)

        if self.type == "hande":
            width = 0.03
        elif self.type == GripperType.Robotiq_2f_85:  # 2f_85
            width = 0.06
        else:
            raise NotImplementedError("Gripper type %s does not have CalcGripperVelocity() implemented" % self.type)

        # Send a single number to match the hardware
        both_finger_velocity = self.ComputeVelocity(state)
        net_velocity = 1 / width * np.mean(both_finger_velocity)

        output.SetFromVector([net_velocity])

    def CalcGripperTorque(self, context, output):
        state = self.state_port.Eval(context)
        target = self.target_port.Eval(context)
        target_type = self.target_type_port.Eval(context)

        finger_position = self.ComputePosition(state)
        finger_velocity = self.ComputeVelocity(state)

        # Set PD gains depending on gripper type
        if self.type == "hande":
            width = 0.03
            Kp = 100 * np.eye(2)
            Kd = 2 * np.sqrt(Kp)
        elif self.type == GripperType.Robotiq_2f_85:
            width = 0.06
            Kp = 10 * np.eye(2)
            Kd = 2 * np.sqrt(0.01 * Kp)
        else:
            raise NotImplementedError("Gripper type %s does not have CalcGripperTorque() implemented" % self.type)

        # Set target positions and velocities based on the current control mode
        if target_type == GripperTarget.kPosition:
            target = width - width * target * np.ones(2)
            target_finger_position = target
            target_finger_velocity = np.zeros(2)
        elif target_type == GripperTarget.kVelocity:
            target_finger_position = finger_position
            target_finger_velocity = -width * target
        else:
            raise RuntimeError("Invalid gripper target type: %s" % target_type)

        # Determine applied torques with PD controller
        position_err = target_finger_position - finger_position
        velocity_err = target_finger_velocity - finger_velocity
        tau = -Kp @ (position_err) - Kd @ (velocity_err)

        output.SetFromVector(tau)



import numpy as np
from pydrake.common.value import AbstractValue
from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import BasicVector

# Internal imports
from brom_drake.control.arms.base_arm_controller import BaseArmController
from brom_drake.control.arms.joint_target import JointTarget


class JointArmController(BaseArmController):
    """
    A controller which imitates the joint control mode of the Kinova gen3 arm.

                                ------------------------
                                |                      |
                                |                      |
    joint_target -------------> |  JointArmController  | ----> applied_arm_torque
    joint_target_type --------> |                      |
                                |                      |
                                |                      | ----> measured_ee_pose
    arm_joint_position -------> |                      | ----> measured_ee_twist
    arm_joint_velocity ------>  |                      |
                                |                      |
                                |                      |
                                ------------------------

    The type of target is determined by ee_target_type, and the options are defined in the
    end effector target.
    """
    def __init__(
        self,
        plant: MultibodyPlant,
        arm_model,
        end_effector_frame_name: str = "end_effector_frame",
        joint_target_type: JointTarget = JointTarget.kPosition,
    ):
        BaseArmController.__init__(
            self,
            plant,
            arm_model,
            end_effector_frame_name=end_effector_frame_name,
        )

        # Define input and Output Ports
        self.joint_target_port, self.joint_target_type_port = None, None
        self.define_joint_target_input_ports(joint_target_type)

        # Define output ports
        self.DeclareVectorOutputPort(
            "applied_arm_torque",
            BasicVector(self.plant.num_actuators()),
            self.CalcArmTorques
        )

    def define_joint_target_input_ports(self, joint_target_type: JointTarget):
        """
        Description
        -----------
        Define the input ports for the JointArmController system
        :return:
        """
        self.joint_target_port = self.DeclareVectorInputPort(
            "joint_target",
            BasicVector(self.plant.num_positions(self.arm)),
        )
        self.joint_target_type_port = self.DeclareAbstractInputPort(
            "joint_target_type",
            AbstractValue.Make(joint_target_type),
        )

    def CalcArmTorques(self, context, output):
        """
        Description
        -----------
        Calculate the arm torques based on the joint target and the current state of the arm.
        :param context:
        :param output:
        :return:
        """
        # Get the current state of the arm
        q = self.arm_joint_position_port.Eval(context)
        qd = self.arm_joint_velocity_port.Eval(context)

        # Set our "internal plant's state" to the current state of the arm
        self.plant.SetPositions(self.context, q)
        self.plant.SetVelocities(self.context, qd)

        # Calculate the torque needed to compensate for gravity
        tau_g = -self.plant.CalcGravityGeneralizedForces(self.context)

        # Indicate what type of command we're recieving
        target_type = self.joint_target_type_port.Eval(context)

        if target_type == JointTarget.kPosition:
            q_des = self.joint_target_port.Eval(context)
            q_err = q_des - q
            qd_des = 0 * q_err
            kp = 3e3
            kd = 2*np.sqrt(3e3)
            tau = kp * q_err + kd * (qd_des - qd) + tau_g

        else:
            raise NotImplementedError(
                f"Only position joint control is currently supported; received {target_type}"
            )

        output.SetFromVector(tau)


"""
Description:
    This file defines an OfflineMotionPlanner class that is used to define the
    offline motion planning role that can be used in Brom scenes.
"""
from brom_drake.scenes.roles import Role, PortPairing

kOfflineMotionPlanner = Role(
    name="OfflineMotionPlanner",
    description="This role is used to define the offline motion planning role.",
    required_input_definitions=[
        PortPairing("goal_position", "goal_position"),
        PortPairing("joint_positions", "joint_positions"),
        PortPairing("joint_position_limits", "joint_position_limits"),
        PortPairing("id", "scene_id"),
    ],
    required_output_definitions=[
        PortPairing("motion_plan", "plan"),
    ],
)



from brom_drake.control.arms.cartesian_arm_controller import CartesianArmController
from brom_drake.control.arms.end_effector_target import EndEffectorTarget
from .gripper_controller import GripperController
from .gripper_target import GripperTarget

__all__ = [
    "CartesianArmController",
    "EndEffectorTarget",
    "GripperTarget",
    "GripperController",
]


from importlib import resources as impresources
import numpy as np
from pydrake.geometry import MeshcatVisualizer, Meshcat
from pydrake.math import RigidTransform
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import AddMultibodyPlantSceneGraph

# Internal Imports
import brom_drake.robots as robots
from brom_drake.scenes import SceneID
from brom_drake.scenes.types.motion_planning import OfflineMotionPlanningScene
from brom_drake.urdf import drakeify_my_urdf


class ShelfPlanningScene(OfflineMotionPlanningScene):
    def __init__(
        self,
        time_step=1e-3,
        use_meshcat: bool = True, # Usually turn off for CI
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.time_step = time_step
        self.use_meshcat = use_meshcat
        self.shelf_position = np.array([-0.4, 0.6, 0.0])

        # If the base link name is not provided,
        # then we will try to do a smart search for it later.
        self.plant, self.scene_graph, self.meshcat = None, None, None

    def add_all_secondary_cast_members_to_builder(self, builder):
        """
        Add all secondary cast members to the builder.
        :param builder: The DiagramBuilder object where we will add all of the secondary cast members.
        :return:
        """
        # Setup
        self.plant, self.scene_graph = AddMultibodyPlantSceneGraph(
            builder,
            time_step=self.time_step,
        )
        self.plant.set_name("ShowMeThisModel_plant")
        self.add_shelf()

        # Connect to Meshcat
        if self.use_meshcat:
            self.meshcat = Meshcat(port=7001)  # Object provides an interface to Meshcat
            m_visualizer = MeshcatVisualizer(self.meshcat)
            m_visualizer.AddToBuilder(builder, self.scene_graph, self.meshcat)

        # Finalize plant and connect it to system
        self.plant.Finalize()

    def add_shelf(self):
        """
        Add the shelf to the scene.
        :return:
        """
        # Setup
        urdf_file_path = str(
            impresources.files(robots) / "models/bookshelf/bookshelf.urdf"
        )

        # Convert the URDF
        new_urdf_path = drakeify_my_urdf(
            urdf_file_path,
            overwrite_old_logs=True,
            log_file_name="bookshelf-conversion.log",
        )

        # Add the shelf to the plant
        model_idcs = Parser(plant=self.plant).AddModels(str(new_urdf_path))
        self.model_idx = model_idcs[0]

        # Weld the bookshelf to the world frame
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName("base_footprint", self.model_idx),
            RigidTransform(
                p=self.shelf_position,
            ),
        )

    @property
    def id(self) -> SceneID:
        return SceneID.kShelfPlanning1


from manipulation.scenarios import AddMultibodyTriad
import numpy as np
from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import LeafSystem, BasicVector


class ShowMeSystem(LeafSystem):
    """
    Description
    -----------
    This function will create a system that will show the user's model.
    :param LeafSystem:
    :return:
    """
    def __init__(
        self,
        plant: MultibodyPlant,
        model_name: str,
        desired_joint_positions: np.ndarray,
        **kwargs,
    ):
        LeafSystem.__init__(self)

        # Constants
        self.plant = plant
        self.model_name = model_name

        # TODO(kwesi): Do we need any ports?
        # Define Input Ports
        n = len(desired_joint_positions)
        self.desired_joint_positions_port = self.DeclareVectorInputPort(
            "desired_joint_positions",
            BasicVector(n),
        )

        # Define Output Ports
        self.DeclareVectorOutputPort(
            "measured_joint_positions",
            BasicVector(n),
            self.SetModelJointPositions,
            {self.time_ticket()}    # indicate that this doesn't depend on any inputs,
        )                           # but should still be updated each timestep

        # Create plant context
        self.mutable_plant_context = None

    def SetModelJointPositions(self, context, output):
        """
        Description
        -----------
        This method will set the joint positions of the model to the desired joint positions.
        :param context:
        :param output:
        :return:
        """
        # Setup

        # Get desired joint positions
        pose_as_vec = self.desired_joint_positions_port.Eval(context)

        # Set the joint positions
        self.plant.SetPositions(
            self.mutable_plant_context,
            pose_as_vec,
        )

        self.plant.SetVelocities(
            self.mutable_plant_context,
            np.zeros(self.plant.num_velocities()),
        )

        output.SetFromVector(pose_as_vec)




from .DrakeReadyURDFConverter import DrakeReadyURDFConverter
from .drakeify import drakeify_my_urdf
from .util import (
    tree_contains_transmission_for_joint,
    create_transmission_element_for_joint,
)

__all__ = [
    'DrakeReadyURDFConverter',
    'drakeify_my_urdf',
    'tree_contains_transmission_for_joint',
    'create_transmission_element_for_joint',
]


import numpy as np
from pydrake.multibody.tree import JacobianWrtVariable
from pydrake.systems.framework import LeafSystem, BasicVector


class EndEffectorWrenchCalculator(LeafSystem):
    """
    A simple system which takes as input joint torques and outputs the corresponding
    wrench applied to the end-effector.

                       ---------------------------------
                       |                               |
                       |                               |
                       |                               |
    joint_positions -> |  EndEffectorWrenchCalculator  | ---> end_effector_wrench
    joint_angles ----> |                               |
    joint_torques ---> |                               |
                       |                               |
                       |                               |
                       ---------------------------------
    """
    def __init__(self, plant, end_effector_frame):
        LeafSystem.__init__(self)

        self.plant = plant
        self.context = self.plant.CreateDefaultContext()
        self.ee_frame = end_effector_frame

        # Inputs are joint positions, angles and torques
        self.q_port = self.DeclareVectorInputPort(
                                "joint_positions",
                                BasicVector(plant.num_positions()))
        self.v_port = self.DeclareVectorInputPort(
                                "joint_velocities",
                                BasicVector(plant.num_velocities()))
        self.tau_port = self.DeclareVectorInputPort(
                                "joint_torques",
                                BasicVector(plant.num_actuators()))

        # Output is applied wrench at the end-effector
        self.DeclareVectorOutputPort(
                "end_effector_wrench",
                BasicVector(6),
                self.CalcEndEffectorWrench)

    def CalcEndEffectorWrench(self, context, output):
        # Gather inputs
        q = self.q_port.Eval(context)
        v = self.v_port.Eval(context)
        tau = self.tau_port.Eval(context)

        # Set internal model state
        self.plant.SetPositions(self.context, q)
        self.plant.SetVelocities(self.context, v)

        # Some dynamics computations
        M = self.plant.CalcMassMatrixViaInverseDynamics(self.context)
        tau_g = -self.plant.CalcGravityGeneralizedForces(self.context)

        # Compute end-effector jacobian
        J = self.plant.CalcJacobianSpatialVelocity(
            self.context,
           JacobianWrtVariable.kV,
           self.ee_frame,
           np.zeros(3),
           self.plant.world_frame(),
           self.plant.world_frame()
        )

        # Compute jacobian pseudoinverse
        Minv = np.linalg.inv(M)
        Lambda = np.linalg.pinv(J@Minv@J.T)
        Jbar = Lambda@J@Minv

        # Compute wrench (spatial force) applied at end-effector
        w = Jbar@(tau-tau_g)

        output.SetFromVector(w)



"""
joint_target.py
Description:

    This file defines the JointTarget enum.
    This class is used to control different manipulators and helps us
    define the type of target that we want to control.
"""
from enum import IntEnum

class JointTarget(IntEnum):
    kPosition = 1
    kVelocity = 2
    kTorque = 3


from enum import Enum

class GripperType(Enum):
    """
    Description
    -----------

    An enum class that defines the type of gripper being used.
    """
    NoGripper = 0
    Robotiq_2f_85 = 1



from brom_drake.scenes.types import BaseScene
from brom_drake.scenes.roles import kOfflineMotionPlanner


class OfflineMotionPlanningScene(BaseScene):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def add_all_secondary_cast_members_to_builder(self, builder):
        pass

    def suggested_roles(self):
        return [kOfflineMotionPlanner]


"""
end_effector_target.py
Description:

    An enum useful for defining the type of target we're giving to a given manipulator.
"""

from enum import Enum


class EndEffectorTarget(Enum):
    kPose = 1   # Pose (position and orientation) expressed as a 7-element vector
                # of the End Effector w.r.t. the base of the arm.
    kTwist = 2
    kWrench = 3



"""
DiagramTarget.py
Description:

    Creates a DiagramTarget class that can be used to monitor the state of a Diagram.
"""
from typing import NamedTuple, List, Union


class DiagramTarget(NamedTuple):
    name: str
    ports: Union[None, List[int]] = None



"""
convenience.py
Description:

    This file contains a couple of convenience functions that are useful for the user of the URDF conversion
    feature.
"""

from typing import Union
from pathlib import Path

# Internal Imports
from .DrakeReadyURDFConverter import DrakeReadyURDFConverter

def drakeify_my_urdf(
    urdf_file_path: Union[str, Path],
    overwrite_old_logs: bool = False,
    log_file_name: str = "conversion.log",
) -> Path:
    """
    Description:

        This function provides a convenience function for the user to convert a URDF file
        into a "Drake-ready" URDF file.
    :param urdf_file:
    :param output_dir:
    :param overwrite_old_logs:
    :param log_file_name:
    :return:
    """
    # Input Processing
    if not isinstance(urdf_file_path, str) and not isinstance(urdf_file_path, Path):
        raise ValueError(
            f"urdf_file_path must be of type str or Path; received {urdf_file_path} of type {type(urdf_file_path)}."
        )

    if isinstance(urdf_file_path, str):
        urdf_file_path = Path(urdf_file_path)

    # Use converter
    converter = DrakeReadyURDFConverter(
        urdf_file_path,
        overwrite_old_logs=overwrite_old_logs,
        log_file_name=log_file_name,
    )

    # Convert the URDF
    new_urdf_path = converter.convert_urdf()

    return new_urdf_path



from typing import List

import numpy as np
from manipulation.scenarios import AddMultibodyTriad
from pydrake.geometry import Meshcat, MeshcatVisualizer
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import AddMultibodyPlantSceneGraph
from pydrake.systems.framework import DiagramBuilder

import xml.etree.ElementTree as ET

from pydrake.systems.primitives import ConstantVectorSource, VectorLogSink

# Internal Imports
from brom_drake.scenes.types import BaseScene
from brom_drake.scenes import SceneID
from brom_drake.scenes.roles.role import Role
from .show_me_system import ShowMeSystem

class ShowMeThisModel(BaseScene):
    def __init__(
        self,
        path_to_model: str,
        desired_joint_positions: List[float],
        base_link_name: str = None,
        time_step: float = 1e3,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Add the model to the scene
        self.path_to_model = path_to_model
        self.q_des = desired_joint_positions
        self.base_link_name = base_link_name
        self.time_step = time_step

        # If the base link name is not provided,
        # then we will try to do a smart search for it later.
        self.plant, self.scene_graph, self.meshcat = None, None, None
        self.model_idx, self.model_name = None, None
        self.show_me_system = None

    def add_all_secondary_cast_members_to_builder(self, builder: DiagramBuilder):
        """
        Description
        -----------
        This method will add just the user's model to the builder.

        :param builder: A DiagramBuilder object from Drake.
        :return:
        """
        # Setup
        self.plant, self.scene_graph = AddMultibodyPlantSceneGraph(
            builder,
            time_step=self.time_step,
        )
        self.plant.set_name("ShowMeThisModel_plant")
        self.meshcat = Meshcat(port=7001)  # Object provides an interface to Meshcat

        # Add Model
        model_idcs = Parser(plant=self.plant).AddModels(self.path_to_model)
        assert len(model_idcs) == 1, \
            f"Only one model should be added in this scene;" + \
            f" received {len(model_idcs)} in the file {self.path_to_model}."
        self.model_idx = model_idcs[0]
        self.model_name = self.plant.GetModelInstanceName(model_idcs[0])

        # Add ShowMeSystem
        self.show_me_system = builder.AddSystem(
            ShowMeSystem(
                plant=self.plant,
                model_name=self.model_name,
                desired_joint_positions=np.array(self.q_des),
            ),
        )

        # Add Source
        desired_joint_positions_source = builder.AddSystem(
            ConstantVectorSource(np.array(self.q_des)),
        )

        # Connect the source to the system
        builder.Connect(
            desired_joint_positions_source.get_output_port(0),
            self.show_me_system.get_input_port(0),
        )

        # Add Sink
        output_joints_sink = builder.AddSystem(
            VectorLogSink(len(self.q_des)),
        )

        # Connect the system to the sink
        builder.Connect(
            self.show_me_system.get_output_port(0),
            output_joints_sink.get_input_port(0),
        )

        # Add A Triad to base?
        AddMultibodyTriad(self.plant.world_frame(), self.scene_graph)

        # Try to collect the base link name if it is not provided
        if self.base_link_name is None:
            self.base_link_name = self.find_a_good_base_link_name()

        # Weld the base link to the world frame
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName(self.base_link_name, self.model_idx),
        )

        # Connect to Meshcat
        m_visualizer = MeshcatVisualizer(self.meshcat)
        m_visualizer.AddToBuilder(builder, self.scene_graph, self.meshcat)

        # Finalize plant and connect it to system
        self.plant.Finalize()

    def find_a_good_base_link_name(self)->str:
        """
        Description
        -----------
        This method will try to find a good base link for the model.
        :return:
        """
        # Setup

        # Parse the model using xml
        if ".urdf" in self.path_to_model:
            original_xml = ET.ElementTree(file=self.path_to_model)
            link_names = self.find_all_link_names(original_xml)

            # Find a link that contains the word "base"
            for link_name in link_names:
                if "base" in link_name.lower():
                    return link_name # Return the first one we find

        else:
            raise ValueError(
                "We can only smartly find base links in .urdf files, for now.\n" +
                "File an issue if you want more support in the future."
            )

        # If we can't find a base link, then we will raise an error
        raise ValueError(
            "We could not find a good base link in the model.\n" +
            "Please provide the base link name manually by adding \"base\" to one of the urdf's links."
        )


    # TODO(kwesi): Maybe move this to a utilities file?
    @staticmethod
    def find_all_link_names(xml_tree: ET.ElementTree)->List[str]:
        """
        Description
        -----------
        This method will find all the link names in the xml tree.
        :param xml_tree: The xml tree that we would like to investigate.
        :return:
        """
        # Setup
        link_names = []

        # Find all the links
        for link in xml_tree.findall(".//link"):
            link_names.append(link.attrib["name"])

        return link_names

    @property
    def id(self):
        return SceneID.kShowMeThisModel

    @property
    def suggested_roles(self) -> List[Role]:
        return []



from pydrake.systems.framework import DiagramBuilder
from typing import List

# Internal Imports
from .constants import Performer

def find_all_systems_with_output_port(
    builder: DiagramBuilder,
    target_port_name: str,
) -> List[Performer]:
    """
    Description
    -----------
    This method looks through the builder and finds all systems that have
    an output port with the given name.
    :param builder:
    :param target_port_name:
    :return:
    """
    # Setup
    potential_performers = []

    # Algorithm
    for system in builder.GetSystems():
        try:
            output_port = system.GetOutputPort(target_port_name)
            potential_performers.append(system)
        finally:
            continue

    return potential_performers

def find_all_systems_with_input_port(
    builder: DiagramBuilder,
    target_port_name: str,
) -> List[Performer]:
    """
    Description
    -----------
    This method looks through the builder and finds all systems that have
    an input port with the given name.
    :param builder:
    :param target_port_name:
    :return:
    """
    # Setup
    potential_performers = []

    # Algorithm
    for system in builder.GetSystems():
        try:
            input_port = system.GetInputPort(target_port_name)
            potential_performers.append(system)
        finally:
            continue

    return potential_performers


from brom_drake.scenes.roles.role import Role, PortPairing
from brom_drake.scenes.roles.offline_motion_planner import kOfflineMotionPlanner

__all__ = [
    "Role",
    "kOfflineMotionPlanner",
    "PortPairing",
]


"""
add_watcher.py
Description:

    This function defines a couple of convenience functions for adding in the watcher
    to a target diagram builder. These functions should be useful for 99% of the
    users of the DiagramWatcher class.
"""
from typing import List, Tuple, Union

import numpy as np
from pydrake.all import DiagramBuilder, Diagram

from brom_drake.DiagramTarget import DiagramTarget
from brom_drake.DiagramWatcher import DiagramWatcher, DEFAULT_PLOT_DIR
from brom_drake.PortWatcher import PortFigureArrangement
from brom_drake.PortWatcher.port_watcher_options import FigureNamingConvention, PortWatcherOptions

PotentialTargetTypes = List[
    Union[
        str,
        Tuple[str, int],
        Tuple[int, int],
        Tuple[str, str],
        Tuple[str, List[int]],
        Tuple[int, List[int]],
        Tuple[str, List[str]],
    ],
]


def add_watcher(
    builder: DiagramBuilder,
    targets: PotentialTargetTypes = None,
    data_dir: str = DEFAULT_PLOT_DIR,
    plot_arrangement: PortFigureArrangement = PortFigureArrangement.OnePlotPerPort,
    figure_naming_convention: FigureNamingConvention = FigureNamingConvention.kFlat,
    file_format: str = "png",
) -> DiagramWatcher:
    """
    Description:

        This function adds a DiagramWatcher to a DiagramBuilder.

    Example Usage:

        watcher = add_watcher(builder)

        watcher = add_watcher(builder, [("plant", 0), ("controller", 0)])

    :param builder: DiagramBuilder. The diagram builder to which we want to add the watcher.
    :param targets: List[Tuple[Union[str, int]]]. The targets that we want to watch.
    :param data_dir: str. The directory in which we will store the data collected by the DiagramWatcher.
    :param plot_arrangement: PortFigureArrangement. The arrangement of the plots.
        (Can be PortFigureArrangement.OnePlotPerPort OR PortFigureArrangement.OnePlotPerDim)
    :param figure_naming_convention: FigureNamingConvention. The naming convention for the figures.
        (Can be FigureNamingConvention.kFlat OR FigureNamingConvention.kHierarchical)
    :return: DiagramWatcher. The watcher that we have added to the diagram builder.
    """
    # Input Processing
    if not isinstance(plot_arrangement, PortFigureArrangement):
        raise ValueError(
            f"plot_arrangement must be of type PortFigureArrangement; received {plot_arrangement} of type" +
            f" {type(plot_arrangement)}."
        )

    # Parse targets list if it exists
    if targets is not None:
        targets = parse_list_of_simplified_targets(builder, targets)

    port_watcher_options = PortWatcherOptions(
        plot_arrangement=plot_arrangement,
        figure_naming_convention=figure_naming_convention,
        file_format=file_format,
    )

    watcher = DiagramWatcher(builder, targets=targets, plot_dir=data_dir, port_watcher_options=port_watcher_options)
    return watcher


def add_watcher_and_build(
    builder: DiagramBuilder,
    targets: PotentialTargetTypes = None,
    data_dir: str = DEFAULT_PLOT_DIR,
    plot_arrangement: PortFigureArrangement = PortFigureArrangement.OnePlotPerPort,
    figure_naming_convention: FigureNamingConvention = FigureNamingConvention.kFlat,
    file_format: str = "png",
) -> (DiagramWatcher, Diagram):
    """
    add_watcher_and_build
    Description:

        This function adds a DiagramWatcher to a DiagramBuilder and then builds the diagram.

    Example Usage:

        watcher = add_watcher(builder)

        watcher = add_watcher(builder, [("plant",)])

        watcher = add_watcher(builder, [("plant", 0), ("controller", 0)])

    :param builder: DiagramBuilder. The diagram builder to which we want to add the watcher.
    :param targets: List[Tuple[Union[str, int]]]. The targets that we want to watch.
    :param data_dir: str. The directory in which we will store the data collected by the DiagramWatcher.
    :param plot_arrangement: PortFigureArrangement. The arrangement of the plots.
        (Can be PortFigureArrangement.OnePlotPerPort OR PortFigureArrangement.OnePlotPerDim)
    :param figure_naming_convention: FigureNamingConvention. The naming convention for the figures.
        (Can be FigureNamingConvention.kFlat OR FigureNamingConvention.kHierarchical)
    :return: DiagramWatcher. The watcher that we have added to the diagram builder.
    """
    watcher = add_watcher(
        builder,
        targets=targets, data_dir=data_dir,
        plot_arrangement=plot_arrangement,
        figure_naming_convention=figure_naming_convention,
        file_format=file_format,
    )

    # Build the diagram and add a reference to the watcher
    diagram = builder.Build()
    diagram_context = diagram.CreateDefaultContext()
    watcher.diagram = diagram
    watcher.diagram_context = diagram_context

    return watcher, diagram, diagram_context


def parse_list_of_simplified_targets(
    builder: DiagramBuilder,
    targets: PotentialTargetTypes,
) -> List[DiagramTarget]:
    """
    parse_list_of_simplified_targets
    Description:

        This function takes a list of simplified targets and converts them to the full form.

    Example Usage:

        targets = [("plant", 0), ("controller", 0)]
        parsed_targets = parse_list_of_simplified_targets(targets)

    :param builder: DiagramBuilder. The diagram builder to which we want to add the watcher.
    :param targets: List[Tuple[Union[str, int]]]. The list of simplified targets.
    :return: List[Tuple[str, int]]. The list of full targets.
    """
    # Setup
    parsed_targets = []
    system_list = builder.GetSystems()

    # If input is not a list, then raise an error
    if not isinstance(targets, list):
        raise ValueError("the input \"targets\" must be a list.")

    if len(targets) == 0:
        raise ValueError(
            "the input \"targets\" must be a non-empty list containing tuples or strings.\n" +
            "If you want to watch the entire diagram, then pass in None as the value of targets."
        )

    # Parse each element in targets
    for target in targets:
        target_name, ports_list = None, None

        if isinstance(target, str):
            target_name = target
        elif isinstance(target, tuple):
            # Parse the first element in the tuple
            if isinstance(target[0], str):
                target_name = target[0]
            elif isinstance(target[0], int):
                # Check if the index is valid
                if target[0] < 0 or target[0] >= len(system_list):
                    raise ValueError(
                        f"the index {target[0]} is not a valid index for the list of systems (has length {len(system_list)})."
                    )
                target_name = system_list[target[0]].get_name()

            else:
                raise ValueError(
                    "the first element of the tuple must be a string or an integer; received {} (type {}).".format(
                        target[0], type(target[0])
                    )
                )

            # Parse the second element in the tuple
            if isinstance(target[1], int):
                ports_list = [target[1]]
            elif isinstance(target[1], str):
                # Find the index of port with this name
                port_name = target[1]

                # Find the system
                system = None
                for ii, system_ii in enumerate(system_list):
                    if system_ii.get_name() == target_name:
                        system = system_ii
                        break

                # for jj in range(system.num_output_ports()):
                #     print(system.get_output_port(jj).get_name())

                if system.HasOutputPort(port_name):
                    ports_list = [int(system.GetOutputPort(port_name).get_index())]
                else:
                    raise ValueError(
                        f"the system {target_name} does not have an output port named {port_name}."
                    )


            elif isinstance(target[1], list):
                ports_list = []
                for ii, elt_ii in enumerate(target[1]):
                    if isinstance(elt_ii, int):
                        ports_list.append(elt_ii)
                    elif isinstance(elt_ii, str):
                        port_name = elt_ii
                        # Find the system
                        system = None
                        for ii, system_ii in enumerate(system_list):
                            if system_ii.get_name() == target_name:
                                system = system_ii
                                break

                        if system.HasOutputPort(port_name):
                            ports_list += [int(system.GetOutputPort(port_name).get_index())]
                        else:
                            raise ValueError(
                                f"the system {target_name} does not have an output port named {port_name}."
                            )

                    else:
                        raise ValueError(
                            f"the target_list[{ii}] of the tuple is not an integer or a string! " +
                            f"Received {elt_ii} of type {type(elt_ii)}."
                        )
            else:
                raise ValueError(
                    "the second element of the tuple must be either a: \n" +
                    "- an integer\n" +
                    "- a list of integers\n" +
                    "- a string\n" +
                    "- a list of strings\n" +
                    "- None.\n" +
                    f"Received type {type(target[1])}"
                )

        # Create the diagram target object
        dt = DiagramTarget(target_name, ports_list)
        parsed_targets.append(dt)

    return parsed_targets



from .port_watcher import PortWatcher
from .port_watcher_options import PortWatcherOptions, FigureNamingConvention
from .port_figure_arrangement import PortFigureArrangement

__all__ = [
    'PortWatcher',
    'PortWatcherOptions',
    'PortFigureArrangement',
    'FigureNamingConvention',
]